use strict;
use warnings;
use File::Basename qw/dirname/;
use IPC::Open3;

my $static_file_dir = dirname(__FILE__) . "/static";
my $tar_command = ["tar", "zcvf", "static.tgz", $static_file_dir];

my $ret = system @$tar_command;
if ( $ret != 0 ) {
  die "archive failed";
}

my $hosts = ["host1", "host2", "host3"];
my $deploy_dir = "~/deploy";
my $timestamp = time;
my $scp_pids = [];
for my $host (@$hosts) {
  my $pid = fork;
  if ($pid == 0) {
    my $ret = system('scp', 'static.tgz', "$host:$deploy_dir");
    exit $ret if $ret != 0;
    $ret = system('ssh', $host, 'mkdir', '-p', "$deploy_dir/$timestamp");
    exit $ret if $ret != 0;
    $ret = system('ssh', $host, 'tar', 'zxvf', "$deploy_dir/static.tgz", '-C', "$deploy_dir/$timestamp");
    exit $ret if $ret != 0;
    exit system('ssh', $host, 'rm', '-f', "$deploy_dir/static.tgz");
  }
  else {
    push(@$scp_pids, $pid);
  }
}

for my $pid (@$scp_pids) {
  waitpid($pid, 0);
  rollback_all($timestamp, $hosts) if $? != 0;
}

my $symlink_pids = [];
for my $host (@$hosts) {
  my $pid = fork;
  if ($pid == 0) {
    exit system('ssh', $host, 'ln', '-nsf', "$deploy_dir/$timestamp", "$deploy_dir/current");
  }
  else {
    push(@$symlink_pids, $pid);
  }
}

for my $pid (@$symlink_pids) {
  waitpid($pid, 0);
  rollback_all($timestamp, $hosts) if $? != 0;
}


sub rollback_all{
  my ($timestamp, $hosts) = @_;
  my $rollback_pids = [];
  for my $host (@$hosts) {
    my $pid = fork;
    if ($pid == 0) {
      rollback($timestamp, $host);
    }
    else {
      push(@$rollback_pids, $pid);
    }
  }
  for my $pid (@$rollback_pids) {
    waitpid($pid, 0);
    die "@@@@!! rollback failed !!@@@@" if $? != 0;
  }
}

sub rollback {
  my ($timestamp, $host)  = @_;

  my $ret = system('ssh', $host, 'rm', '-rf', "$deploy_dir/current");
  exit $ret if $ret != 0;

  $ret = system('ssh', $host, 'rm', '-rf', "$deploy_dir/$timestamp");
  exit $ret if $ret != 0;

  my $pid = open3(my $wtr, my $rdr, undef, 'ssh', $host, 'ls', "$deploy_dir");
  close $wtr;

  waitpid($pid, 0);
  exit $? if $? != 0;

  my @files = <$rdr>;
  close $rdr;

  @files = sort {$b <=> $a} @files;
  my $latest_file = shift @files;
  chomp $latest_file;

  exit system('ssh', $host, 'ln', '-nsf', "$deploy_dir/$latest_file", "$deploy_dir/current");
}
